/*
 * rrlms.c - Road Runner login program
 *
 * 6-16-98
 *
 * Larkin Lowrey - llowrey@NuclearWinter.com
 * http://Larkin.NuclearWinter.com/rros2/
 *
 * This is freeware.
 *
 * See http://Larkin.NuclearWinter.com/rros2/Digital.pdf for information
 * on the Digital authentication protocol
*/

/*
 * mods by George Coulouris <glc5@cornell.edu>
 * http://www.tc.cornell.edu/~glc5/
*/

/*
 * more mods by Nanda Santhanam <nanda@lightlink.com>
 * for:
 * - daemonization
 * - logout protocol
 * - catch other signals
 * - get other ports from server
 * - syslogging
 * - '-k' flag to kill running server
 * - add usage blurb
 * - misc. clean-up
*/

/*		Don Wurster [5/15/99]
 * Hacked for use as external login/logout functions for a graphical Roadrunner
 * app for BeOS
 * - removed almost everything that makes it good command line
 * - turned main into rrlmsLogin
 * - turned Logout into rrlmsLogout
*/

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/ioctl.h>
//#include <sys/errno.h>
#include <pwd.h>
#include <stdio.h>
#include <unistd.h>
#include "global.h"
#include "md5.h"
#include <string.h>

/* This implements an experimental Digital authentication RR login */

char InPacket[1024], OutPacket[1024];
char OS_Name[20], OS_Version[20];
unsigned int AuthAddr, ProtocolAddr;
unsigned short AuthPort = 60001, LogoutPort = 60002, StatusPort = 60003;
unsigned short VersionID;
char Username[65], Password[65], Server[65];

int Lookup(char *name, unsigned int *ip)
{
  struct hostent *he;

  if (!ip)
    return -1;

  alarm(30);
  he=gethostbyname(name);
  alarm(0);
  if (!he)
    return -2;
	
  *ip=*(int *)he->h_addr_list[0];

  return 0;
}

int NewPacket(short type)
{
  *(short *)OutPacket=htons(type);
  *(short *)(OutPacket+2)=htons(8);
  *(short *)(OutPacket+4)=0;
  *(short *)(OutPacket+6)=0;

  return 0;
}

int NewParam(short type, char *buffer, short len)
{
  short packetlen;

  packetlen=ntohs(*(short *)(OutPacket+2));

  *(short *)(OutPacket+packetlen)=htons(type);
  memcpy(OutPacket+packetlen+4,buffer,len);
  *(short *)(OutPacket+packetlen+2)=htons(len+4);
  *(short *)(OutPacket+2)=htons(packetlen+len+4);

  return 0;
}

int GetData(short type, unsigned char *buffer)
{
  unsigned char *c, *max;

  c = (unsigned char *)InPacket+8;
  max=c+ntohs(*(short *)(InPacket+2))-8;
  while (*(short *)c!=htons(type) && c<max)
    c+=ntohs(*(short *)(c+2));
	
  if (c>=max)
    return -1;

  memcpy(buffer,c+4,ntohs(*(short *)(c+2))-4);
  return 0;
}

int GetString(short type, unsigned char *buffer)
{
  unsigned char *c, *max;

  c=(unsigned char *)InPacket+8;
  max=c+ntohs(*(short *)(InPacket+2))-8;
  while (*(short *)c!=htons(type) && c<max)
    c+=ntohs(*(short *)(c+2));
	
  if (c>=max)
    return -1;

  memcpy(buffer,c+4,ntohs(*(short *)(c+2))-4);
  buffer[ntohs(*(short *)(c+2))-4]=0;
  return 0;
}

void
Hash(unsigned char *target, unsigned char *source, int len)
{
  md5_state_t context;
  unsigned char *c;
  int len1;

  c=source;
  len1=len;
  //PrintLog("Hashing 0x");
  while (len1--)
  {
    //PrintLog("%02X",*c);
    c++;
  }
  //PrintLog("\n\n");


  memset((char *)&context,0,sizeof(context));
  md5_init(&context);
  md5_append(&context,source,len);
  md5_finish(&context, target);
}

int ProtocolNegotiation(char *server)
{
  int sock,rc;
  struct sockaddr_in sin;
  unsigned short num;
  char szAuthAddr[64];

  if (Lookup(server,&ProtocolAddr) != 0)
  {
    //LogSysMessage(LOG_ERR, "Unable to resolve IP for protocol-server\n");
    return -1;
  }

  if ((sock=socket(AF_INET,SOCK_STREAM,0))==-1)
  {
    //LogSysMessage(LOG_ERR, "Failed to allocate socket\n");
    return -2;
  }

  sin.sin_family=AF_INET;
  sin.sin_port=htons(60000);
  sin.sin_addr.s_addr=ProtocolAddr;

  if (connect(sock,(struct sockaddr *)&sin,sizeof(sin))==-1)
  {
    //LogSysMessage(LOG_ERR, "Unable to connect to server\n");
    close(sock);
    return -3;
  }

  NewPacket(1);
  NewParam(3,(char *)&VersionID,sizeof(VersionID));
  NewParam(4,OS_Name,strlen(OS_Name));
  NewParam(5,OS_Version,strlen(OS_Version));
  num=htons(1);
  NewParam(1,(char *)&num,sizeof(num));

  //PrintPacket(OutPacket);
  rc=send(sock,OutPacket,ntohs(*(short *)(OutPacket+2)),0);
  //PrintLog("Send RC=%d\n",rc);
  if (rc==-1)
  {
    close(sock);
    return -4;
  }

  rc=recv(sock,InPacket,1024,0);
  //PrintLog("Received... RC=%d\n",rc);
  //PrintRawPacket(InPacket,rc);
  //PrintPacket(InPacket);
  if (rc==-1)
  {
    close(sock);
    return -5;
  }

  if (*(short *)InPacket!=htons(2))
  {
    //LogSysMessage(LOG_ERR, "Unexpected packet type from server\n");
    close(sock);
    return -6;
  }

  if (GetData(10,(unsigned char *)&num)!=0 || num)
  {
    //LogSysMessage(LOG_ERR, "Invalid response from server\n");
    close(sock);
    return -7;
  }

  GetString(24,(unsigned char *)szAuthAddr);
  GetData(15,(unsigned char *)&AuthPort);

  //LogSysMessage(LOG_INFO,
  //              "Authentication server %s:%d\n",szAuthAddr,ntohs(AuthPort));

  //PrintLog("lookup %s\n",szAuthAddr);
  if (Lookup(szAuthAddr,&AuthAddr)!=0)
  {
    //LogSysMessage(LOG_ERR,
    //              "Authentication host, %s, could not be located\n",szAuthAddr);
    return -8;
  }

  close(sock);

  return 0;
}





/*
 ok.. these are going to be the two external functions called from RRLogin
 get ready for nasty hacks..

*/

int rrlmsLogoff( void )
/*
 Commented out all calls to LogSysMessage as the returned int value will
 indicate the error for RRLogin to pop-up
 
 Commented out calls to PrintLog.. since this isnt going to be running in 
 a term anymore..
*/
{
  unsigned short num, status;
  struct sockaddr_in sin;
  int sock, rc;
  unsigned int blinding;
  unsigned char *uc;
  char rText[256];
  int ret;
  unsigned short hashPwd;
  unsigned char nonce[16];
  char hashed[16], hashsrc[1024];


  ret = 0;
  sock = -1;

  /* open socket */
  if ( (sock=socket(AF_INET,SOCK_STREAM,0) )==-1)
  {
    //LogSysMessage(LOG_ERR, "Failed to allocate socket\n");
    ret = -1;
  }

  if ( ret == 0 )
  {
    sin.sin_family = AF_INET;
    sin.sin_port = LogoutPort;
    sin.sin_addr.s_addr = AuthAddr;
    uc = (unsigned char *)&sin.sin_addr.s_addr;
    //LogSysMessage(LOG_INFO,
    //              "Connecting for logout to: %d.%d.%d.%d:",
    //              uc[0],uc[1],uc[2],uc[3]);
    uc=(unsigned char *)&sin.sin_port;
    //PrintLog("%d\n",ntohs(*(short *)uc));
    if ( connect(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0 )
    {
      //LogSysMessage( LOG_ERR,
      //               "Unable to connect to auth server for logout\n" );
      ret = -2;
    }
  }

  if ( ret == 0 )
  {
    /* send logout request */
    NewPacket( 6 );
    NewParam( 7, Username, strlen(Username) );
    NewParam( 3, (char *)&VersionID, sizeof(VersionID) );
    NewParam( 4, OS_Name, strlen(OS_Name) );
    NewParam( 5, OS_Version, strlen(OS_Version) );
    num = 0;
    NewParam( 6, (char *)&num, sizeof(num) );
    
    //PrintLog( "Sending...\n" );
    //PrintPacket( OutPacket );
    rc = send( sock, OutPacket, ntohs(*(short *)(OutPacket+2)), 0 );
    //PrintLog( "SendRC=%d\n", rc );

    /* get challenge back */
    if ( (rc = recv(sock, InPacket, 1024, 0)) < 8 )
    {
      //LogSysMessage( LOG_ERR,
      //               "Error receiving from auth server; rc = %d\n", rc );
      ret = -3;
    }
  }

  if ( ret == 0 )
  {
    //PrintLog( "Received...rc=%d\n", rc );
    //PrintRawPacket( InPacket, rc );
    //PrintPacket( InPacket );
    if ( *(short *)InPacket != htons(9) )
    {
      //LogSysMessage( LOG_ERR,
      //               "Packet type not expected\n" );
      ret = -4;
    }
  }

  if ( ret == 0 )
  {
    GetData( 0x000E, (unsigned char *)&hashPwd );
    GetData( 0x000C, nonce );
    
    blinding = htonl( time(NULL) );
    if ( hashPwd )
    {
      memcpy( hashsrc,nonce,16 );
      Hash((unsigned char *)hashsrc+16,(unsigned char *)Password,strlen(Password) );
      memcpy( hashsrc+16+16,(char *)&blinding,4 );
      hashsrc[16+16+4]=0;
      hashsrc[16+16+4+1]=7;
      memset( hashed,0,16 );
      Hash( (unsigned char *)hashed,(unsigned char *)hashsrc,16+16+4+2 );
    }
    else
    {
      memcpy( hashsrc,nonce,16 );
      strcpy( hashsrc+16,Password );
      memcpy( hashsrc+16+strlen(Password),(char *)&blinding,4 );
      hashsrc[16+strlen(Password)+4]=0;
      hashsrc[16+strlen(Password)+4+1]=7;
      Hash( (unsigned char *)hashed,(unsigned char *)hashsrc,16+strlen(Password)+4+2 );
    }

    NewPacket( 7 );
    NewParam( 11, hashed, 16 );
    NewParam( 21, (char *)&blinding, 4 );
    
    //PrintLog( "Sending...\n" );
    //PrintPacket( OutPacket );
    rc=send(sock,OutPacket,ntohs(*(short *)(OutPacket+2)),0);
    //PrintLog( "SendRC = %d\n", rc );
  
    
    if ( (rc=recv(sock,InPacket,1024,0)) < 0 )
    {
      //LogSysMessage( LOG_ERR,
      //               "Error receiving from auth server; rc=%d\n",rc );
      ret = -3;
    }
  }

  if ( ret == 0 )
  {
    //PrintLog( "Received...rc = %d\n", rc );
    //PrintRawPacket( InPacket, rc );
    //PrintPacket( InPacket );
    GetData( 10, (unsigned char *)&status );
    if ( status == 0 )
      //LogSysMessage( LOG_INFO, "Logout successful\n" );
	  ret = 1;
    else
    {
      GetData( 9, (unsigned char *)rText );
      //LogSysMessage( LOG_ERR,
      //               "Logout failed; status: %u\nError message: %s\n",
      //				ntohs(status), rText );
      ret = -5;
    }
  }

  if ( sock > 0 )
    close( sock );
    
  return( ret );
}



int rrlmsLogin( void )
/*
  Commented out pid, command line args, and loging stuff
  
  Changed all exit() calls into return()s for error pop-ups
  
  Commented out while(1) section... err.. It works with out it,
  but ill have to read up on what it did..
*/
{

  short num;
  struct sockaddr_in sin;
  int sock, rc, sinlen, kasock, sequence;
  unsigned int blinding;
  unsigned short  kaport, status;
  unsigned char *uc;
  struct tm *tm;
   
  //pid_t mypid;
  FILE *f;
  char hashed[16], hashsrc[1024];
  unsigned short hashPwd;
  unsigned char nonce[16];

  /* open syslog */
  //openlog( PROG_NAME, LOG_PID, LOG_LOCAL0 );

//  Verbose = 0;

//  /* check whether if server needs to be killed */
//  if ( argc > 1 )
//  {
//    if ( strcmp(argv[1], "-k") == 0 )
//    {
//      int error = 0;
//      FILE *pidFile = NULL;
//      unsigned long pid;
//      
//      pidFile = fopen( PID_FILE, "r" );
//      if ( pidFile == NULL )
//      {
//        LogSysMessage( LOG_ERR,
//                       "%s file does not exist.  Server may not be running\n",
//                       PID_FILE );
//        error = -1;
//      }
//      if ( error == 0  && fscanf(pidFile, "%lu", &pid) != 1 )
//      {
//	LogSysMessage( LOG_ERR, "Cannot read pid from %s\n", PID_FILE );
//	error = -1;
//      }
//      if ( error == 0 && kill((pid_t)pid, SIGHUP) != 0 )
//	LogSysMessage( LOG_ERR, "Cannot kill the server\n" );
//      
//      if ( pidFile )
//	fclose( pidFile );
//      
//      exit(error);
//    }
//    else
//    {
//      Usage( argv[0] );
//      exit( -1 );
//    }
//  }
   
  VersionID=htons(11);
  strcpy(OS_Name,"Windows 95");
  strcpy(OS_Version,"4.0");

  f=fopen("/boot/beos/etc/rrlms.conf","r");
   
  if (f == NULL)
  {
    //LogSysMessage(LOG_ERR,"Couldn't open configuration file: %s\n",CONF_FILE);
    return(-6);
  }
  else
  {
    fscanf(f,"%64s %64s %64s",Server, Username, Password);
    /* fprintf(stderr,"Server: %s, Username: %s, Password: %s\n",Server, Username,Password); */
    fclose(f);
  }
   
//  f=fopen(PID_FILE,"r");
//  if (f != NULL)
//  {
//    LogSysMessage(LOG_ERR,"%s exists. another instance running?\n", PID_FILE);
//    fclose(f);
//    exit(1);
//  }
      
//  /* daemonize */
//  if ( (mypid = fork()) < 0 )
//  {
//    /* cannot fork */
//    LogSysMessage(LOG_ERR, "Cannot fork. Exiting\n");
//    exit(-1);
//  }
//  else if ( mypid > 0 )
//  {
//    /* parent */
//    f=fopen(PID_FILE,"w");
//    fprintf(f,"%d\n",mypid);
//    fclose(f);
//    exit(0);
//  }
//
//  signal(SIGHUP, signal_handler);
//  signal(SIGINT, signal_handler);
//  signal(SIGTERM, signal_handler);
   
  if (ProtocolNegotiation(Server)!=0)
  {
    //LogSysMessage(LOG_ERR, "Protocol Negotiation failed. Unable to proceed\n");
    //remove(PID_FILE);
    return(-1);
  }

  if ((kasock=socket(AF_INET,SOCK_DGRAM,0))==-1)
  {
    //LogSysMessage(LOG_ERR, "Unable to allocate Keep-Alive socket\n");
    //remove(PID_FILE);
    return(-1);
  }
	
  sin.sin_family=AF_INET;
  sin.sin_port=0;
  sin.sin_addr.s_addr=0;

  bind(kasock,(struct sockaddr *)&sin,sizeof(sin));

  sinlen=sizeof(sin);
  getsockname(kasock,(struct sockaddr *)&sin,&sinlen);
  kaport=sin.sin_port;

  if ((sock=socket(AF_INET,SOCK_STREAM,0))==-1)
  {
    //LogSysMessage(LOG_ERR, "Unable to allocate socket\n");
    //remove(PID_FILE);
    return(-1);
  }

  sin.sin_family=AF_INET;
  sin.sin_port=AuthPort;
  sin.sin_addr.s_addr=AuthAddr;

  uc=(unsigned char *)&sin.sin_addr.s_addr;
  //PrintLog("Connecting to: %d.%d.%d.%d:",uc[0],uc[1],uc[2],uc[3]);
  uc=(unsigned char *)&sin.sin_port;
  //PrintLog("%d\n",ntohs(*(short *)uc));
  if (connect(sock,(struct sockaddr *)&sin,sizeof(sin))<0)
  {
    //LogSysMessage(LOG_ERR, "Unable to connect to auth server\n");
    //remove(PID_FILE);
    return(-2);
  }

  NewPacket(3);
  NewParam(7,Username,strlen(Username));
  NewParam(3,(char *)&VersionID,sizeof(VersionID));
  NewParam(4,OS_Name,strlen(OS_Name));
  NewParam(5,OS_Version,strlen(OS_Version));
  num=0;
  NewParam(6,(char *)&num,sizeof(num));
  NewParam(8,(char *)&kaport,sizeof(kaport));

  //PrintLog("Sending...\n");
  //PrintPacket(OutPacket);
  rc=send(sock,OutPacket,ntohs(*(short *)(OutPacket+2)),0);
  //PrintLog("SendRC=%d\n",rc);

  if ((rc=recv(sock,InPacket,1024,0))<8)
  {
    //LogSysMessage(LOG_ERR, "Error receiving from auth server; rc=%d\n",rc);
    //remove(PID_FILE);
    return(-3);
  }
  //PrintLog("Received...rc=%d\n",rc);
  //PrintRawPacket(InPacket,rc);
  //PrintPacket(InPacket);

  if (*(short *)InPacket!=htons(9))
  {
    //LogSysMessage(LOG_ERR, "Packet type not expected\n");
    //remove(PID_FILE);
    return(-4);
  }

  GetData(0x000E,(unsigned char *)&hashPwd);
  GetData(0x000C,nonce);

  blinding=htonl(time(NULL));

  if (hashPwd)
  {
    memcpy(hashsrc,nonce,16);
    Hash((unsigned char *)hashsrc+16,(unsigned char *)Password,strlen(Password));
    memcpy(hashsrc+16+16,(char *)&blinding,4);
    hashsrc[16+16+4]=0;
    hashsrc[16+16+4+1]=4;
    memset(hashed,0,16);
    Hash((unsigned char *)hashed,(unsigned char *)hashsrc,16+16+4+2);
  }
  else
  {
    memcpy(hashsrc,nonce,16);
    strcpy(hashsrc+16,Password);
    memcpy(hashsrc+16+strlen(Password),(char *)&blinding,4);
    hashsrc[16+strlen(Password)+4]=0;
    hashsrc[16+strlen(Password)+4+1]=4;
    Hash((unsigned char *)hashed,(unsigned char *)hashsrc,16+strlen(Password)+4+2);
  }

  NewPacket(4);
  NewParam(11,hashed,16);
  NewParam(21,(char *)&blinding,4);

	
  //PrintLog("Sending...\n");
  //PrintPacket(OutPacket);
  rc=send(sock,OutPacket,ntohs(*(short *)(OutPacket+2)),0);
  //PrintLog("SendRC=%d\n",rc);

  if ((rc=recv(sock,InPacket,1024,0))<0)
  {
    //LogSysMessage(LOG_ERR, "Error receiving from auth server; rc=%d\n",rc);
    //remove(PID_FILE);
    return(-3);
  }
  close(sock);
  //PrintLog("Received...rc=%d\n",rc);
  //PrintRawPacket(InPacket,rc);
  //PrintPacket(InPacket);

  if (*(short *)InPacket!=htons(5))
  {
    //LogSysMessage(LOG_ERR, "unexpected response\n");
    //remove(PID_FILE);
    return(-5);
  }

  /* check login return code */
  GetData( 10, (unsigned char *)&status );
  status = ntohs( status );
  if ( status == 0 || status == 100 || status == 102 )
  {
//LogSysMessage( LOG_INFO, "Login successful.  Status is %u\n",
    //               (unsigned int) status );
  }
  else
  {
    //LogSysMessage( LOG_ERR, "Login failed.  Error is %u\n",
    //               (unsigned int) status );
    //remove(PID_FILE);
    return(-8);
  }

  GetData(16,(unsigned char *)&LogoutPort);
  GetData(17,(unsigned char *)&StatusPort);
  //LogSysMessage(LOG_INFO,
  //              "Logout port is: %u\n", (unsigned int)ntohs(LogoutPort));
  //LogSysMessage(LOG_INFO,
  //              "Status port is: %u\n", (unsigned int)ntohs(StatusPort));

  rc=time(NULL);
  tm=localtime((const time_t *)&rc);
  //LogSysMessage(LOG_INFO, "Logged in @ %s",asctime(tm));

  //LogSysMessage(LOG_INFO, "Keep-alive UDP port %d\n",ntohs(kaport));

  sequence=0;
  //while(1)
//  {
//    sinlen=sizeof(sin);
//    rc=recvfrom(kasock,InPacket,1024,0,(struct sockaddr *)&sin,&sinlen);
//    now=time(NULL);
//    tm=localtime((const time_t *)&now);
//    //LogSysMessage(LOG_INFO,
//    //              "Received %d bytes from %s:%d\n @ %s",rc,inet_ntoa(sin.sin_addr),ntohs(sin.sin_port),asctime(tm));
//    //PrintRawPacket(InPacket,rc);
//    //PrintPacket(InPacket);
//
//
//    if (*(short *)InPacket!=htons(11))
//    {
//      //LogSysMessage(LOG_ERR, "Unexpected packet\n");
//      continue;
//    }
//
//    NewPacket(12);
//    num=0;
//    NewParam(10,(unsigned char *)&num,sizeof(num));
//
//    sequence++;
//    bignum=htonl(sequence);
//    if (hashPwd)
//    {
//      memcpy(hashsrc,nonce,16);
//      Hash(hashsrc+16,Password,strlen(Password));
//      memcpy(hashsrc+16+16,(char *)&bignum,4);
//      hashsrc[16+16+4]=0;
//      hashsrc[16+16+4+1]=12;
//      memset(hashed,0,16);
//      Hash(hashed,hashsrc,16+16+4+2);
//    }
//    else
//    {
//      memcpy(hashsrc,nonce,16);
//      strcpy(hashsrc+16,Password);
//      memcpy(hashsrc+16+strlen(Password),(char *)&bignum,4);
//      hashsrc[16+strlen(Password)+4]=0;
//      hashsrc[16+strlen(Password)+4+1]=12;
//      Hash(hashed,hashsrc,16+strlen(Password)+4+2);
//    }
//    NewParam(19,hashed,16);
//    NewParam(13,(unsigned char *)&bignum,4);
//
//    //PrintLog("Sending...\n");
//    //PrintPacket(OutPacket);
//    sin.sin_port=StatusPort;
//    rc=sendto(kasock,OutPacket,ntohs(*(short *)(OutPacket+2)),0,(struct sockaddr *)&sin,sizeof(sin));
//    //PrintLog("SendRC=%d\n",rc);
//  }

return(2);
}
